#include <stdio.h>
#include <string.h>
#include <windows.h>

#ifndef _MSC_VER
#error Tested on MSVC++ 6 only!
#endif

#ifndef MAX_PATH
#define MAX_PATH 1024
#endif

#ifdef SELF_EXE
#define WORKING_CP  CP_OEMCP
#else
#define WORKING_CP  CP_ACP
#endif

/* UTF8.c */
wchar_t *decode_UTF8(wchar_t *ws, char *utf8s);

/* already defined in windows.h */
//typedef unsigned __int8     BYTE;
//typedef unsigned __int16    WORD;
//typedef unsigned __int32    DWORD;
//typedef unsigned __int32    LONG;
typedef unsigned __int64    QWORD;
typedef __int64             LLONG;

typedef struct _GI_UD {          /* Global Unique ID */
    DWORD   dw;
    WORD    w[2];
    BYTE    b[8];
} GU_ID;

typedef struct _hdrsec_info {   /* header section info */
    LLONG   off;                /* Offset of section from beginning of file */
    LLONG   len;                /* Length of section */
} hdrsec_info;

typedef struct _chm_filehdr {
    char        magic[4];       /* "ITSF" */
    DWORD       ver;            /* Version number */
    DWORD       totlen;         /* Total header length, including header section table and following data */
    DWORD       uk1;            /* 1 (unknown) */
    DWORD       ts;             /* timestamp */
    DWORD       LangID;         /* Windows Language ID */
    GU_ID       GUIDs[2];
    hdrsec_info fsi;
    hdrsec_info dsi;
} chm_filehdr;

#define CHM_FILEHDR_MAGIC   ("ITSF")

typedef struct _chm_fsizeinfo {
    DWORD   magic;              /* $01FE (unknown) */
    DWORD   uk1;
    LLONG   fsize;              /* File Size */
    DWORD   uk3;
    DWORD   uk4;
} chm_fsizeinfo;

#define CHM_FSIZEINFO_MAGIC 0x01FE

typedef struct _chm_dirhdr {    /* Directory header */
    char    magic[4];           /* "ITSP" */
    DWORD   ver;                /* Version number 1 */
    DWORD   len1;               /* Length of the directory header */
    DWORD   uk1;                /* $0a (unknown) */
    DWORD   dircs;              /* $1000 Directory chunk size */
    DWORD   qrden;              /* "Density" of quickref section, usually 2 */
    DWORD   itdep;              /* Depth of the index tree 1 there is no index, 2 if there is one level of PMGI chunks */
    LONG    cnri;               /* Chunk number of root index chunk, -1 if there is none
                                   (though at least one file has 0 despite there being no index chunk, probably a bug.) */
    DWORD   cn1PMGL;            /* Chunk number of first PMGL (listing) chunk */
    DWORD   cnlPMGL;            /* Chunk number of last PMGL (listing) chunk */
    DWORD   uk2;                /* -1 (unknown) */
    DWORD   cnum;               /* Number of directory chunks (total) */
    DWORD   LangID;             /* Windows language ID */
    GU_ID   aGUID;              /* {5D02926A-212E-11D0-9DF9-00A0C922E6EC} */
    DWORD   len2;               /* 0x54 (This is the length again) */
    LONG    uk3;                /* -1 (unknown) */
    LONG    uk4;                /* -1 (unknown) */
    LONG    uk5;                /* -1 (unknown) */
} chm_dirhdr;

#define CHM_DIRHDR_MAGIC    ("ITSP")

typedef struct _chm_dirchunk {  /* The Listing Chunks */
    char    magic[4];           /* "PMGL" */
    DWORD   qrlen;              /* Length of free space and/or quickref area at end of directory chunk */
    DWORD   uk1;                /* Always 0 (???) */
    LONG    prev;               /* Chunk number of previous listing chunk when reading
                                   directory in sequence (-1 if this is the first listing chunk) */
    LONG    next;               /* Chunk number of next listing chunk when reading
                                   directory in sequence (-1 if this is the last listing chunk) */
} chm_dirchunk;

#define CHM_DIRCHUNK_MAGIC  ("PMGL")


/* Matthew T. Russotto (from chmformat.html):
 *
 * An ENCINT is a variable-length integer.
 * The high bit of each byte indicates "continued to the next byte". Bytes
 * are stored most significant to least significant.
 * So, for example, $EA $15 is (((0xEA & 0x7F) << 7 ) | 0x15) = 0x3515.
 *
 * Reading can be stopped by 0, not NOT "continued to the next byte"
 */
DWORD read_encint(FILE *fp)
{
    DWORD   val = 0;
    int     i = 7, c;

    //printf("-|");
    while((c = fgetc(fp)) != EOF) {
        //printf(" %02x", c);
        val += c & 0x7f;
        if(c & 0x80) {
            val <<= 7;
            //i += 7;
        }
        else break;
    }
    //printf("|-");
    return val;
}

char *GU_ID_2_str(char *buf, GU_ID g)
{
    if(buf) {
        sprintf(buf, "%08lX-%04X-%04X-%02X%02X-%02X%02X-%02X%02X-%02X%02X",
                g.dw, g.w[0], g.w[1],
                g.b[0], g.b[1], g.b[2], g.b[3], g.b[4], g.b[5], g.b[6], g.b[7]);
    }
    return buf;
}

void main(int argc, char *argv[])
{
    FILE            *chm = NULL;
    char            *fnchm, buf[1024];
    wchar_t         wbuf[1024];
    chm_filehdr     cfh;
    chm_fsizeinfo   fsi;
    chm_dirhdr      dsi;
    chm_dirchunk    dch;
    DWORD           firstdc = 0, afsec, afoff, aflen, cstart;
    BYTE            fnlen;
    int             i, chkcnt = 0, filecnt = 0;

    if(argc != 2) {
        printf("Usage: chmlist <chmfile>\n");
        goto stop;
    }
    fnchm = argv[1];
    if(!(chm = fopen(fnchm, "rb"))) {
        printf("Can not open %s\n", fnchm);
        goto stop;
    }
    printf("File header:\n");
    fread(&cfh, sizeof(chm_filehdr), 1, chm);
    if(strncmp(cfh.magic, CHM_FILEHDR_MAGIC, strlen(CHM_FILEHDR_MAGIC))) {
        printf("Magic (%.4s) not correct. Must: %s\n", cfh.magic, CHM_FILEHDR_MAGIC);
        goto stop;
    }
    printf("Version: %u, total len: %u (%X), uk1: %u, time: %u, LangID: %#06x\n",
            cfh.ver, cfh.totlen, cfh.totlen, cfh.uk1, cfh.ts, cfh.LangID);
    printf("GUID 1: %s\n", GU_ID_2_str(buf, cfh.GUIDs[0]));
    printf("GUID 2: %s\n", GU_ID_2_str(buf, cfh.GUIDs[1]));
    printf("File size info off: %I64d (%I64X), len: %I64d (%I64X)\n", cfh.fsi.off, cfh.fsi.off, cfh.fsi.len, cfh.fsi.len);
    printf("Dir section off: %I64d (%I64X), len: %I64d (%I64X)\n", cfh.dsi.off, cfh.dsi.off, cfh.dsi.len, cfh.dsi.len);

    fseek(chm, cfh.fsi.off, SEEK_SET);
    printf("%06X: File size info:\n", ftell(chm));
    if(!fread(&fsi, sizeof(chm_fsizeinfo), 1, chm)) {
        printf("Can not read\n");
        goto stop;
    }
    if(fsi.magic != CHM_FSIZEINFO_MAGIC) {
        printf("Magic (%04X) not correct. Must: %04X\n", fsi.magic, CHM_FSIZEINFO_MAGIC);
        goto stop;
    }
    printf("File size: %I64d (%I64X)\n", fsi.fsize, fsi.fsize);

    fseek(chm, cfh.dsi.off, SEEK_SET);
    printf("%06X: Directory section info (%u (%X)):\n", ftell(chm), sizeof(chm_dirhdr), sizeof(chm_dirhdr));
    if(!fread(&dsi, sizeof(chm_dirhdr), 1, chm)) {
        printf("Can not read\n");
        goto stop;
    }
    if(strncmp(dsi.magic, CHM_DIRHDR_MAGIC, strlen(CHM_DIRHDR_MAGIC))) {
        printf("Magic (%.4s) not correct. Must: %s\n", fsi.magic, CHM_FSIZEINFO_MAGIC);
        goto stop;
    }
    firstdc = ftell(chm);
    printf( "Version: %u, header len: %u (%X), uk1: %u (%X), chunk size: %u (%X)\n"
            "qrden: %u, itdep: %u, cnri: %d\n"
            "1st PMGL: %u, last PMGL: %u, chunks num: %u\n"
            "LangID: %#06x, GUID: %s, len2: %u (%X)\n"
            "uk2: %X, uk3: %X, uk4: %X\n"
            "dirstart: %06X\n",
            dsi.ver, dsi.len1, dsi.len1, dsi.uk1, dsi.uk1, dsi.dircs, dsi.dircs,
            dsi.qrden, dsi.itdep, dsi.cnri,
            dsi.cn1PMGL, dsi.cnlPMGL, dsi.cnum, dsi.LangID, GU_ID_2_str(buf, dsi.aGUID),
            dsi.len2, dsi.len2,
            dsi.uk2, dsi.uk3, dsi.uk4,
            firstdc);

    printf("Reading directory chunks...\n", ftell(chm));
    for(chkcnt = 0, filecnt = 0, dch.next = dsi.cn1PMGL; -1 != dch.next; ) {
        fseek(chm, firstdc + dch.next * dsi.dircs, SEEK_SET);
        cstart = ftell(chm);
        printf("%06X: chunk: %2u (%2X)", cstart, dch.next, dch.next);
        fread(&dch, sizeof(chm_dirchunk), 1, chm);
        printf(". Magic \"%.4s\" ", dch.magic);
        if(strncmp(dch.magic, CHM_DIRCHUNK_MAGIC, strlen(CHM_DIRCHUNK_MAGIC))) {
            printf(". not correct. Must: %s\n", CHM_DIRCHUNK_MAGIC);
            goto stop;
        }
        ++chkcnt;
        printf("OK, prev: %2d (%2X), next: %2d (%2X), qrlen: %u (%X)\n",
                dch.prev, dch.prev, dch.next, dch.next, dch.qrlen, dch.qrlen);
        do {
            printf("%06X: ", ftell(chm)); fflush(stdout);
            //fread(&fnlen, sizeof(BYTE), 1, chm);
            fnlen = read_encint(chm);
            fread(buf, sizeof(BYTE), fnlen, chm);
            buf[fnlen] = 0;
            decode_UTF8(wbuf, buf);
            WideCharToMultiByte(WORKING_CP, 0, wbuf, -1, buf, MAX_PATH, NULL, NULL);
            printf("%s", buf); fflush(stdout);
            afsec = read_encint(chm);
            printf(" s:%d", afsec); fflush(stdout);
            afoff = read_encint(chm);
            printf(" o:%d (%X)", afoff, afoff); fflush(stdout);
            aflen = read_encint(chm);
            printf(" l:%d (%X)\n", aflen, aflen); fflush(stdout);
            ++filecnt;
        } while(ftell(chm) < (cstart + dsi.dircs - dch.qrlen));
    }
    printf("Total chunks: %d, files: %d\n", chkcnt, filecnt);

stop:
    if(chm) fclose(chm);
}
