#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <direct.h>
#include <shellapi.h>
#include "wcxhead.h"
#include "chmlib.h"
#include "chmdir.h"

// from hhc2html.c
extern int hhc2html(char *sfn, char *dfn, char *fhref);
extern int comp_str(char *s1, char *s2);

//#define DEBUG
#ifdef DEBUG
    static char log_name[_MAX_PATH] = "";
    static FILE *log;
#endif

#ifdef DEBUG
static char *make_fullpath(char *buf, char *bp, char *sp)
/* Adds non full name sp from sp and stores to buf.
 * Returns buf, if OK, else NULL (buf stays unchanged).
 */
{
    char bpd[_MAX_DRIVE], bpdir[_MAX_DIR], bpn[_MAX_FNAME], bpe[_MAX_EXT];
    char spd[_MAX_DRIVE], spdir[_MAX_DIR], spn[_MAX_FNAME], spe[_MAX_EXT];

    if(!buf || !bp || !sp) return 0;
    _splitpath(bp, bpd, bpdir, bpn, bpe);
    _splitpath(sp, spd, spdir, spn, spe);
    if(!strlen(spd)) strcpy(spd, bpd);
    if(!strlen(spdir)) strcpy(spdir, bpdir);
    if(!strlen(spn)) strcpy(spn, bpn);
    if(!strlen(spe)) strcpy(spe, bpe);
    _makepath(buf, spd, spdir, spn, spe);
    return buf;
}
#endif

static int del_dir(char *d)
/* Deletes dir with all subdirs.
 * Returns 0 if OK, else !0.
 */
{
    struct  _finddata_t fi;
    long    h;
    int     rc = 0, drc = 0;
    char    odir[_MAX_PATH];

    if(!getcwd(odir, sizeof(odir))) {
    #ifdef DEBUG
        if(log = fopen(log_name, "a")) {
            fprintf(log, "Can get cwd!\n");
            fclose(log);
        }
    #endif
        return 1;
    }
    if(chdir(d)) {
    #ifdef DEBUG
        if(log = fopen(log_name, "a")) {
            fprintf(log, "Can not change to %s\n", d);
            fclose(log);
        }
    #endif
        return 1;
    }
    rc = h = _findfirst("*.*", &fi);
    while(rc != -1) {
        if(fi.attrib & _A_SUBDIR) {
            if(*fi.name != '.') {
                drc = del_dir(fi.name);
            #ifdef DEBUG
                if(log = fopen(log_name, "a")) {
                    fprintf(log, "Deleting dir %s %s\n", fi.name, drc ? "Error" : "Ok");
                    fclose(log);
                }
            #endif
            }
        }
        else {
            drc = remove(fi.name);
        #ifdef DEBUG
            if(log = fopen(log_name, "a")) {
                fprintf(log, "Removing file %s %s\n", fi.name, drc ? "Error" : "Ok");
                fclose(log);
            }
        #endif
        }
        rc = _findnext(h, &fi);
    }
    _findclose(h);
    if(chdir(odir)) {
    #ifdef DEBUG
        if(log = fopen(log_name, "a")) {
            fprintf(log, "Can not change to %s\n", d);
            fclose(log);
        }
    #endif
        return 1;
    }
    rc = rmdir(d);
#ifdef DEBUG
    if(log = fopen(log_name, "a")) {
        fprintf(log, "Removing dir %s %s (odir: %s)\n", d, rc ? "Error" : "Ok", odir);
        fclose(log);
    }
#endif
    return rc;
}

static int make_dir(char *fn)
/* Make dir for file fn if it not exists.
 * Returns 0 if OK, else 1.
 */
{
    char fpath[_MAX_PATH], drv[_MAX_DRIVE], dir[_MAX_DIR],
            nam[_MAX_FNAME], ext[_MAX_EXT], *s;
    struct _finddata_t fi;
    long h;
    int rc = 0;

    _splitpath(fn, drv, dir, nam, ext);
    _makepath(fpath, drv, dir, "", "");
    s = fpath + strlen(fpath) - 1;
    if(*s == '\\') *s = 0;
    h = _findfirst(fn, &fi);
    if(h != -1) {
        _findclose(h);
        if(fi.attrib & _A_SUBDIR) rc = 1;
    }
    else rc = mkdir(fpath) == 0;
    return rc;
}

static int x_chm_file(char *fchm, char *d)
/* Unpacks file with name fchm into dir d.
 * d must ends by '\'. If *d == 0 uses current dir.
 * Returns 0 if OK, else !0.
 */
{
    HANDLE              h;
    tOpenArchiveData    ad;
    tHeaderData         hd;
    int                 hrc, prc;
    char                fpath[_MAX_PATH], *s;

    ad.ArcName = fchm;
    ad.OpenMode = PK_OM_EXTRACT;
    if(!(h = OpenArchive(&ad))) {
    #ifdef DEBUG
        getcwd(fpath, sizeof(fpath));
        if(log = fopen(log_name, "a")) {
            fprintf(log, "Can not open: \"%s\", cwd: \"%s\". OpenResult: %d\n",
                    fchm, fpath, ad.OpenResult);
            fclose(log);
        }
    #endif
        return 1;
    }
    do {
        hrc = ReadHeader(h, &hd);
        if(hrc != E_END_ARCHIVE) {
            strcpy(fpath, d);
            s = fpath + strlen(fpath) - 1;
            if(*s != ':' && *s != '\\') strcat(s, "\\");
            strcat(s, hd.FileName);
            make_dir(fpath);
            prc = ProcessFile(h, PK_EXTRACT, d, hd.FileName);
        }
    #ifdef DEBUG
        if(log = fopen(log_name, "a")) {
            fprintf(log, "Dir: %s, File: %s %s\n", d, hd.FileName, prc ? "Error!" : "Ok.");
            fclose(log);
        }
    #endif
    } while(hrc != E_END_ARCHIVE);
    CloseArchive(h);
    return 0;
}

static char *ch_slash(char *s)
{
    char *ts = s;
    while(*s) { if(*s == '\\') *s = '/'; ++s; }
    return ts;
}

static char *find_1file(char *d, char *fm, char *buf)
/* Finds in dir d and it's subdirs 1st file matched to fm.
 * If found its path will be written into buf.
 * buf must large enough.
 * fm may be NULL. Mask will be "*.*".
 * d may be NULL. Search in current dir.
 *
 * Returns buf, if OK, else NULL.
 */
{
    struct _finddata_t fi;
    long    h;
    int     rc = 0, found = 0;
    char    fspec[_MAX_PATH] = "";

    if(!buf) return NULL;
    if(fm) _makepath(fspec, NULL, d, fm, NULL);
    else   _makepath(fspec, NULL, d, "*.*", NULL);

    // first match in current dir
    rc = h = _findfirst(fspec, &fi);
    while(rc != -1 && !found) {
        if(!(fi.attrib & _A_SUBDIR)) {
            _makepath(buf, NULL, d, fi.name, NULL);
            found = 1;
        }
        if(!found) rc = _findnext(h, &fi);
    }
    _findclose(h);
    if(found) return buf;

    // search subdirs and then in these subdirs
    _makepath(fspec, NULL, d, "*.*", NULL);
    rc = h = _findfirst(fspec, &fi);
    while(rc != -1 && !found) {
        if((fi.attrib & _A_SUBDIR) && (fi.name[0] != '.')) {
            _makepath(fspec, NULL, d, fi.name, NULL);
            found = find_1file(fspec, fm, buf) != NULL;
        }
        if(!found) rc = _findnext(h, &fi);
    }
    _findclose(h);
    if(found) return buf;

    return NULL;
}

static char *make_wdir(char *chmn, char *wdir)
{
    char *s, drv[_MAX_DRIVE], dir[_MAX_DIR], n[_MAX_FNAME], e[_MAX_EXT];

    _splitpath(chmn, drv, dir, n, e);
    if(!(s = getenv("temp"))) s = getenv("tmp");
    _makepath(wdir, NULL, s, n, "chmv");
    del_dir(wdir);
    if(mkdir(wdir)) return NULL;
    return wdir;
}

static int parse_arg(char *cmdline, int argn, char *buf, int buflen)
{
    char    *s, *ts, delim;
    int     i;

    if(argn < 1 || !buf) {
        return 1;
    }
    else {
        for(s = cmdline; *s && *s <= ' ' && *s != '\"'; ) ++s;
        if(!*s) return 1;
        i = 1;
        do {
            delim = (*s == '\"') ? *s : ' ';
            if(delim == '\"') ++s;
            ts = s;
            if(delim == '\"')  while(*s && *s != delim) ++s;
            else               while(*s && *s > ' ') ++s;
            if(argn == i) {
                if(s-ts > (buflen-1)) return 1;
                strncpy(buf, ts, s-ts);
                buf[s-ts] = 0;
                return 0;
            }
            if(delim == '\"' && *s) ++s;
            while(*s && *s <= ' ' && *s != '\"') ++s;
            ++i;
        } while(*s);
    }
    return 1;
}

static int count_args(char *cmdline)
{
    char    *s, delim;
    int     i = 0;

    for(s = cmdline; *s && *s <= ' ' && *s != '\"'; ) ++s;
    if(*s) {
        do {
            delim = (*s == '\"') ? *s : ' ';
            if(delim == '\"') ++s;
            if(*s) {
                ++i;
                if(delim == '\"') {
                    while(*s && *s != delim) ++s;
                    if(*s) ++s;
                }
                else {
                    while(*s && *s > ' ') ++s;
                }
            }
            while(*s && *s <= ' ' && *s != '\"') ++s;
        } while(*s);
    }
    return i;
}

int WINAPI WinMain(
        HINSTANCE hInstance,        // handle to current instance
        HINSTANCE hPrevInstance,    // handle to previous instance
        LPSTR lpCmdLine,            // pointer to command line
        int nCmdShow                // show state of window
)
/*   --   .chm
 */
{
    char    buf[_MAX_PATH+512], buf1[256], wdir[_MAX_PATH] = "", *s, chmn[_MAX_PATH+2];
    char    modname[_MAX_PATH], hhcn[_MAX_PATH], fhref[_MAX_PATH] = "";
    char    contn[] = "__chmv__.html", frfn[] = "__chmvf__.html";
    int     rc = 0;
    FILE    *ff;

    // init CHMDir
    chmdir_init(GetModuleHandle(__argv[0]));

    GetModuleFileName(NULL, modname, _MAX_PATH);

#ifdef DEBUG
    make_fullpath(log_name, modname, "chmview.log");
    if(log = fopen(log_name, "w")) {
        fprintf(log, "Log file name: %s\n", log_name);
        fclose(log);
    }
#endif

    if(count_args(lpCmdLine) != 1) {
        MessageBox(NULL, "You must specify 1 argument: CHM file name", "CHMView: command line",
                MB_OK | MB_ICONINFORMATION);
        return 1;
    }
    // name of CHM and working dir
    parse_arg(lpCmdLine, 1, chmn, _MAX_PATH);
    if(!make_wdir(chmn, wdir)) {
        MessageBox(NULL, "Can not make working dir", "CHMView: Error",
                MB_OK | MB_ICONERROR);
        return 1;
    }
#ifdef DEBUG
    if(log = fopen(log_name, "a")) {
        fprintf(log, "Temp dir: %s\nCHM file: %s\n", wdir, chmn);
        fclose(log);
    }
#endif
    if(x_chm_file(chmn, wdir)) {
        del_dir(wdir);
        sprintf(buf, "Can not extract \"%s\" to \"%s\"", chmn, wdir);
        MessageBox(NULL, buf, "CHMView: Error", MB_OK | MB_ICONERROR);
        return 1;
    }
#ifdef DEBUG
    //MessageBox(NULL, "CHM extracted", "CHMView: info", MB_OK);
#endif
    buf[0] = 0;
    chdir(wdir);
    if(find_1file(NULL, "*.hhc", hhcn)) {
        if(rc = hhc2html(hhcn, contn, fhref)) {
            sprintf(buf, "Error while translating .hhc file.\n");
            //MessageBox(NULL, buf, "CHMView: error", MB_OK | MB_ICONERROR);
            return 1;
        }
        if(!(ff = fopen(frfn, "w"))) {
            sprintf(buf, "Can not open %s.\n", frfn);
            //MessageBox(NULL, buf, "CHMView: error", MB_OK | MB_ICONERROR);
            return 1;
        }
        fprintf(ff, "<HTML><HEAD><title>%s</title></HEAD>\n", ch_slash(chmn));
        fprintf(ff, "<FRAMESET cols=\"25%%,*\">\n"
                "<FRAME SRC=\"%s\">\n"
                "<FRAME SRC=\"%s\" NAME=\"main_text\">\n"
                "</FRAMESET>", ch_slash(contn), ch_slash(fhref));
        fprintf(ff, "<NOFRAMES><a href=\"%s\">View without frames</a>"
                "</NOFRAMES></HTML>", ch_slash(contn));
        fclose(ff);
        ShellExecute(NULL, 0, frfn, NULL, NULL, SW_SHOWNORMAL);
    }
    else if(find_1file(NULL, "index.htm*", hhcn)) {
        ShellExecute(NULL, 0, hhcn, NULL, NULL, SW_SHOWNORMAL);
    }
    else if(find_1file(NULL, "*index.htm*", hhcn)) {
        ShellExecute(NULL, 0, hhcn, NULL, NULL, SW_SHOWNORMAL);
    }
    else {
        ShellExecute(NULL, 0, wdir, NULL, NULL, SW_SHOWNORMAL);
        sprintf(buf, "Can not find any index file. Try find it manually.\n");
        //MessageBox(NULL, buf, "CHMDir: no index file", MB_OK | MB_ICONINFORMATION);
    }
    sprintf(buf+strlen(buf),
            "Delete directory \"%s\"\n"
            "containing unpacked files from\"%s\"?", wdir, chmn);
    rc = MessageBox(NULL, buf, "CHMView: delete directory?", MB_YESNO | MB_ICONQUESTION);
    if(rc == IDYES) {
        chdir("..");
        del_dir(wdir);
    }
    return 0;
}
