#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* Mnemonics of last 64 symbols in 8-bit encoding.
 * <mnemonic>, <dec_code>, <symbol>
 */
static char *hi_latin[] = {
    "Agrave", "192", "",
    "Aacute", "193", "",
    "Acirc" , "194", "",
    "Atilde", "195", "",
    "Auml"  , "196", "",
    "Aring" , "197", "",
    "AElig" , "198", "",
    "Ccedil", "199", "",
    "Egrave", "200", "",
    "Eacute", "201", "",
    "Ecirc" , "202", "",
    "Euml"  , "203", "",
    "Igrave", "204", "",
    "Iacute", "205", "",
    "Icirc" , "206", "",
    "Iuml"  , "207", "",
    "ETH"   , "208", "",
    "Dstrok", "208", "",
    "Ntilde", "209", "",
    "Ograve", "210", "",
    "Oacute", "211", "",
    "Ocirc" , "212", "",
    "Otilde", "213", "",
    "Ouml"  , "214", "",
    "times" , "215", "",
    "Oslash", "216", "",
    "Ugrave", "217", "",
    "Uacute", "218", "",
    "Ucirc" , "219", "",
    "Uuml"  , "220", "",
    "Yacute", "221", "",
    "THORN" , "222", "",
    "szlig" , "223", "",
    "agrave", "224", "",
    "aacute", "225", "",
    "acirc" , "226", "",
    "atilde", "227", "",
    "auml"  , "228", "",
    "aring" , "229", "",
    "aelig" , "230", "",
    "ccedil", "231", "",
    "egrave", "232", "",
    "eacute", "233", "",
    "ecirc" , "234", "",
    "euml"  , "235", "",
    "igrave", "236", "",
    "iacute", "237", "",
    "icirc" , "238", "",
    "iuml"  , "239", "",
    "eth"   , "240", "",
    "ntilde", "241", "",
    "ograve", "242", "",
    "oacute", "243", "",
    "ocirc" , "244", "",
    "otilde", "245", "",
    "ouml"  , "246", "",
    "divide", "247", "",
    "oslash", "248", "",
    "ugrave", "249", "",
    "uacute", "250", "",
    "ucirc" , "251", "",
    "uuml"  , "252", "",
    "yacute", "253", "",
    "thorn" , "254", "",
    "yuml"  , "255", ""
};
static int hi_latin_sz = sizeof(hi_latin) / sizeof(char*);

extern int comp_str(char *s1, char *s2)
/* Compares to strings ignoring case (1st half of codepage).
 * Returns <0 if s1<s2, >0 if s1>s2 and 0 if strings are equal.
 */
{
    int rc = 0;

    while((*s1 || *s2) && !rc) {
        rc = toupper(*s1) - toupper(*s2);
        ++s1;
        ++s2;
    }
    return rc;
}

static char *parse_par(char *pars, char *parn, char *buf)
/* Searches in string pars that looks like
 *   param1="value1" param2=value2
 * parameter with name parn (ignoring case). If parameter found, writes
 * its value into buf. Quotation marks discarded.
 * If parameter not found buf unchanged.
 *
 * Returns buf if OK else NULL.
 */
{
    char    *s, *bp, tpar[32] = "", *rp = NULL;

    s = pars;
    while(sscanf(s, " %[^=]", tpar) != EOF) {
        //printf("s: %s, tpar: %s\n", s, tpar);
        /* s will point to start of value */
        while(*s && *s != '=') ++s;
        ++s;
        if(comp_str(tpar, parn)) {
            /* tag not matched, skip value */
            if(*s == '\"') {
                ++s;
                while(*s && *s != '\"') ++s;
                ++s;
            }
            else {
                while(*s && !strchr(" \n\r\t\f\v", *s)) ++s;
            }
        }
        else {
            /* matched, copy value */
            if(*s == '\"') {
                ++s;
                sscanf(s, "%[^\"]", buf);
            }
            else {
                bp = buf;
                while(*s && !strchr(" \n\r\t\f\v", *s)) *bp++ = *s++;
            }
            rp = buf;
            break;
        }
    }
    return rp;
}

static char *get_tagpars(FILE *fp, char *buf)
/* Writes tag parameters starting current position in file *fp into buf.
 * Returns buf.
 */
{
    int     c, i;
    char    *s, ent[32];

    s = buf;
    while((c = fgetc(fp)) != EOF && c != '>') {
        if(c == '&') {
            /* test on chars with code >0x7f (mnemonic) */
            fscanf(fp, "%[^;]", ent);
            c = fgetc(fp);
            for(i = 0; i < hi_latin_sz; i += 3) {
                if(!strcmp(hi_latin[i], ent)) break;
            }
            if(i < hi_latin_sz) strcpy(s, hi_latin[i+2]);
            else sprintf(s, "&%s;", ent);
            while(*s) ++s;
        }
        else {
            *s = c;
            ++s;
        }
    }
    *s = 0;
    return buf;
}

static int parse_object(FILE *fp, char *href, char *atext)
/* Gets from tag <OBJECT> in file *fp values of parameters
 * with name "Name" (anchor text) and "Local" (anchor link).
 * Name --> atext, Local --> href.
 *
 * Returns 0 if OK else !0 (href and atext will be empty).
 */
{
    int     rc = 0, c;
    char    tagname[64] = "", *tagpars;
    char    objtype[64], parname[64];

    *href = 0;
    *atext = 0;
    tagpars = malloc(1024);
    if(!tagpars) {
        rc = 1;
        goto stop;
    }
    get_tagpars(fp, tagpars);
    if(!parse_par(tagpars, "type", objtype) || comp_str(objtype, "text/sitemap")) {
        /* type of object not specified or it is not "text/sitemap".
         * Ignoring all up to </OBJECT> inclusive */
        while((c = fgetc(fp)) != EOF) {
            if(c == '<') {
                fscanf(fp, "%[^> ]", tagname);
                get_tagpars(fp, tagpars);
                if(!comp_str(tagname, "/OBJECT")) break;
            }
        }
        rc = 1;
        goto stop;
    }

    /* work with this object */
    while(comp_str(tagname, "/OBJECT")) {
        /* skip all up to tag */
        while((c = fgetc(fp)) != EOF && c != '<') ;
        fscanf(fp, "%[^> ]", tagname);
        get_tagpars(fp, tagpars);
        if(!comp_str(tagname, "param") && parse_par(tagpars, "name", parname)) {
            if(!comp_str(parname, "Name"))
                    parse_par(tagpars, "value", atext);
            else if(!comp_str(parname, "Local"))
                    parse_par(tagpars, "value", href);
        }
    }
stop:
    if(tagpars) free(tagpars);
    return rc;
}

extern int hhc2html(char *sfn, char *dfn, char *fhref)
/* Converts contents of CHM (.hhc) with name sfn into HTML file with name
 * dfn, replacing tag <OBJECT> with parameters "Name" and "Local" by tag <A>,
 * where "Name" -- anchor text and "Local" -- link.
 *
 * Example:
 *
 *  <OBJECT type="text/sitemap">
 *       <param name="Name" value="FlashGet Help">
 *       <param name="Local" value="index.htm">
 *       <param name="ImageNumber" value="1">
 *       </OBJECT>
 *
 * will be replaced by
 *
 *  <A HREF="index.htm">FlashGet Help</A>
 *
 * BTW mnemonics of high half of encoding table (codes from 192 to 255)
 * will be replaced by corresponding symbols, e.g. "Atilde" to "".
 *
 * 1st link will be written into fhref.
 *
 * Returns 0 if OK else !0.
 */
{
    FILE    *sf, *df;
    int     rc = 0;
    char    tagname[64], tagpars[128];
    char    atext[512], href[256];
    int     c, ul_nest = 0;
    char    *s;

    sf = fopen(sfn, "r");
    df = fopen(dfn, "w");
    if(!sf || !df) {
        rc = 1;
        goto stop;
    }
    *fhref = 0;
    while((c = fgetc(sf)) != EOF) {
        if(c != '<') fputc(c, df);
        else { /* there is tag here */
            fscanf(sf, "%[^> ]", tagname);
            if(!comp_str(tagname, "OBJECT")) {
                if(!parse_object(sf, href, atext)) {
                    if(*href == '/') {
                        s = href + 1;
                        while(*s) {
                            *(s-1) = *s;
                            ++s;
                        }
                        *(s-1) = 0;
                    }
                    if(*href)
                        fprintf(df, "<a href=\"%s\" target=main_text>%s</a>\n",
                            href, atext);
                    else
                        fprintf(df, "%s\n", atext);
                    if(fhref && !*fhref && *href) strcpy(fhref, href);
                }
            }
            else {
                get_tagpars(sf, tagpars);
                if(!comp_str(tagname, "HEAD")) {
                    fprintf(df, "<%s%s>", tagname, tagpars);
                    fprintf(df, "<style>li {margin-left: -1em;} a {font-size: smaller;}</style>");
                }
                else if(!comp_str(tagname, "UL")) {
                    if(ul_nest) fprintf(df, "<%s%s>", tagname, tagpars);
                    ++ul_nest;
                }
                else if(!comp_str(tagname, "/UL")) {
                    --ul_nest;
                    if(ul_nest) fprintf(df, "<%s%s>", tagname, tagpars);
                }
                else if(!comp_str(tagname, "LI")) {
                    if(ul_nest > 1) fprintf(df, "<%s%s>", tagname, tagpars);
                }
                else {
                    fprintf(df, "<%s%s>", tagname, tagpars);
                }
            }
        }
    }

stop:
    if(sf) fclose(sf);
    if(df) fclose(df);
    return rc;
}

//#define DEBUG

#ifdef DEBUG

void main(void)
{
    char fhref[128];

    hhc2html("fg_toc.hhc", "fg_toc.htm", fhref);
    hhc2html("regindex.hhc", "regindex.htm", fhref);
}

#endif
